"""
used for managing user Accounts
(Server)

author: Nilusink
"""
from contextlib import suppress
from typing import List
import random
import json
import os


class Manager:
    """
    account manager
    """

    def __init__(self, account_file: str) -> None:
        """
        account_file - file to store encrypted account data in
        """
        self.__encryptionFile = account_file

        self.__accounts = self.__get_file()

    @property
    def names(self) -> List[str]:
        return [user["Name"] for user in self.__accounts]

    def __get_file(self) -> list:
        """
        decrypt the user passwords with threads
        """
        return json.load(open(self.__encryptionFile))

    def __write_accounts(self, accounts: list) -> None:
        """
        write account file
        """
        self.__accounts = accounts
        with open(self.__encryptionFile, 'w') as out:
            json.dump(self.__accounts, out, indent=4)

    def get_accounts(self) -> list:
        """
        get account data
        """
        return self.__accounts

    def set_pwd(self, username: str, new_password: str) -> None:
        """
        set password of given user
        """
        for element in self.__accounts:
            if element['Name'] == username:
                element['pwd'] = new_password  # if user is selected user, change its password
                break  # to not further iterate all users

        self.__write_accounts(self.__accounts)  # write output to file

    def set_username(self, old_user: str, new_user: str) -> None:
        """
        change username
        """
        used_names = self.names
        used_names.remove(old_user)

        element = str()
        i = int()

        # name+'2' because the double-vote agent uses this for their votes
        if new_user not in used_names + [name + '2' for name in used_names]:
            for i, element in enumerate(self.__accounts):
                if element['Name'] == old_user:
                    element['Name'] = new_user  # if user is selected user, change its password
                    continue

            self.__accounts[i] = element  # make sure the new element is in list and on correct position

            self.__write_accounts(self.__accounts)  # write output to file
            return
        raise NameError('Username already exists')

    def set_user_sec(self, username: str, security_clearance: str) -> None:
        """
        set clearance of user
        """
        element = str()
        i = int()

        for i, element in enumerate(self.__accounts):
            if element['Name'] == username:
                element['sec'] = security_clearance  # if user is selected user, change its security clearance
                continue

        self.__accounts[i] = element  # make sure the new element is in list and on correct position
        self.__write_accounts(self.__accounts)

    def new_user(self, username: str, password: str, security_clearance: str) -> None:
        """
        add new user
        """
        # variables
        new_acc: dict
        ids: list
        new_id: int

        if username in self.names:
            raise NameError(f'Username {username} already exists: {self.names}')

        # create new id for user
        ids = [user["id"] for user in self.__accounts]
        new_id = random.randint(10000, 99999)

        if len(ids) >= 90000:
            raise NotImplementedError("Too many accounts registered!")

        while new_id in ids:
            new_id = random.randint(10000, 99999)

        # add account
        new_acc = {
            'Name': username,
            'pwd': password,
            'sec': security_clearance,
            "id": new_id
        }

        self.__accounts.append(new_acc)  # create user

    def remove_user(self, username: str) -> None:
        """
        remove a user
        """
        accounts = self.get_accounts()  # get accounts
        for i in range(len(accounts)):  # iterate accounts
            if accounts[i]['Name'] == username:  # if account name is username
                accounts.pop(i)  # remove user
                break

        self.__write_accounts(accounts)  # update accounts

    def verify(self, username: str, password: str) -> (None | bool, dict):
        """
        return False or user security Clearance
        """
        users = self.get_accounts()  # get accounts
        auth = False
        user = {}
        for element in users:  # iterate users
            if username == element['Name'] and password == element['pwd']:  # if username is account name
                if 'sec' in element:
                    if element['sec'] == '':
                        auth = None
                        continue
                    user = element
                    auth = True
                    break

        return auth, user  # return result


# user configuration
USER_CONFIG: dict = {}
try:
    USER_CONFIG = json.load(open(os.getcwd() + '/config/user_config.json', 'r'))

except FileNotFoundError:
    with suppress(FileNotFoundError):
        USER_CONFIG = json.load(open('/home/apps/Fridrich/config/user_config.json', 'r'))



"""
All fridrich errors

Author:
Nilusink
"""


###########################################################################
#                              Error Classes                              #
###########################################################################
class ServerError(Exception):
    pass


class AccessError(Exception):
    pass


class AuthError(Exception):
    pass


class JsonError(Exception):
    pass


class NoVotes(Exception):
    pass


class UnknownError(Exception):
    pass


class RegistryError(Exception):
    pass


class NotAUser(Exception):
    pass


class InvalidRequest(Exception):
    pass


class SecurityClearanceNotSet(Exception):
    pass


class MessageError(Exception):
    pass


class InvalidStringError(Exception):
    pass


class NetworkError(Exception):
    pass


class Error(Exception):
    pass



"""
used to interface with a Fridrich Server
(Client)

Author: Nilusink
"""
from concurrent.futures import ThreadPoolExecutor, Future
from typing import Dict, Iterable, List, Any, Union
from traceback import format_exc
from contextlib import suppress
from hashlib import sha512
import typing
import socket
import struct
import json
import time
import os

# fridrich imports
from fridrich.file_transfer import send_receive, download_program as now_download_program
from fridrich.file_transfer import download_progress as now_download_progress
from fridrich.backend.debugging import Debugger
from fridrich.classes import Daytime
from fridrich import cryption_tools
from fridrich.errors import *
from fridrich import *


############################################################################
#                     global Variable definition                           #
############################################################################
# Protocol version information
COMM_PROTOCOL_VERSION = "1.1.1"

# debugging
DEBUGGER = Debugger(os.getcwd()+"/logs/fridrich.err", file_mode="a")


############################################################################
#                      Server Communication Class                          #
############################################################################
class FridrichFuture:
    def __init__(self) -> None:
        self.__value: typing.Any | None = ...

    @property
    def result(self) -> typing.Any:
        if self.__value is not ...:
            return self.__value
        raise ValueError("No value received yet")

    @result.setter
    def result(self, value) -> None:
        self.__value = value

    def __repr__(self) -> str:
        return f"<future: result={self.__value is not ...}>"

    def __bool__(self) -> bool:
        """
        return True if AuthKey
        """
        return self.__value is not ...


class Connection:
    def __init__(self, debug_mode: str | None = False, host: str | None = ...) -> None:
        """
        connect with any fridrich server

        :param debug_mode: "normal" | "full" | False
        :param host: name of the host, either IP or hostname / address
        """
        self._messages = dict()
        self._server_messages = dict()
        self.Server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  # create socket instance
        self._debug_mode = debug_mode

        self.__ServerIp = None

        if host is not ...:
            self.server_ip = host

        if self._debug_mode in ('normal', 'full'):
            print(ConsoleColors.OKGREEN + 'Server IP: ' + self.server_ip + ConsoleColors.ENDC)
        self.port = 12345   # set communication port with server

        self.__AuthKey = None
        self.__userN = None
        self.__pwd_hash = None

        self.loop = True

        self.executor = ThreadPoolExecutor(max_workers=1)
        self.receive_thread = Future()

        # for down-/uploading
        self.load_state = str()
        self.load_progress = float()
        self.load_program = str()

        # message pool
        self.__message_pool: list = []
        self.__results_getters: Dict[str, FridrichFuture] = {}

        # optimization variables
        self.__server_time_delta = None
        self.__server_voting_time = None

        self.__login_time = None

    # properties
    @property
    def username(self) -> str:
        """
        get username
        """
        return self.__userN

    @property
    def server_ip(self) -> str:
        return self.__ServerIp

    @server_ip.setter
    def server_ip(self, value: str) -> None:
        sl = value.split('.')
        if len(sl) == 4 and all([digit in '0123456789' for element in sl for digit in element]):
            if self._debug_mode in ("full", "normal"):
                try:
                    socket.gethostbyaddr(value)
                except socket.herror:
                    print(ConsoleColors.WARNING+f"Hostname of {value} not found, may be unreachable"+ConsoleColors.ENDC)
            self.__ServerIp = value
        else:
            self.__ServerIp = socket.gethostbyname(value)  # get ip of fridrich

        if self._debug_mode == 'full':
            print(self.server_ip)

    @property
    def debug_mode(self) -> str:
        return self._debug_mode

    @debug_mode.setter
    def debug_mode(self, value: str) -> None:
        allowed = ("normal", "full", False)
        if value not in allowed:
            raise ValueError(f"must be {' or '.join([str(el) for el in allowed])}")
        self._debug_mode = value

    @property
    def login_time(self) -> Daytime | None:
        return self.__login_time

    # "local" functions
    @staticmethod
    def error_handler(error: str, *args) -> None:
        """
        Handle incoming errors
        """
        match error:    # match errors where not specific error class exists (and NotEncryptedError)
            case 'NotVoted':
                raise NameError('No votes registered for user')

            case 'json':
                raise JsonError('Crippled message')

            case 'SecurityNotSet':
                raise SecurityClearanceNotSet(args[0]['info'])

            case 'NotEncryptedError':
                raise cryption_tools.NotEncryptedError('Server received non encrypted message')

            case _:  # for all other errors try to raise them and when that fails, raise a ServerError
                if 'full' in args[0] and 'info' in args[0]:
                    st = f'raise {error}("{args[0]["info"]} -- Full Traceback: {args[0]["full"]}")'

                elif 'info' in args[0]:
                    st = f'raise {error}("{args[0]["info"]}")'

                else:
                    st = f'raise {error}'

                try:
                    exec(st)

                except NameError:
                    raise ServerError(f'{error}:\n{st.lstrip(f"raise {error}(").rstrip(")")}')

    @DEBUGGER.catch_and_write(raise_error=True, print_traceback=True)
    def response_handler(self, responses: Dict[str, Any]) -> Dict[str, Any]:
        """
        if necessary, process each result
        """
        if self._debug_mode == "full":
            print(f"handling responses {responses=}")
        for response in responses.keys():
            match response:
                case "getFrees":
                    responses[response] = responses[response]['Value']

                case "getChat":
                    raw = responses[response]
                    responses[response] = sorted(raw, key=date_for_sort)

                case "get_var":
                    responses[response] = responses[response]["var"]

                case "ping":
                    responses[response] = (time.time() - responses["ping"]["time"]) * 1000

                case "get_time":
                    self.__server_voting_time = Daytime.from_strftime(responses[response]["voting"])
                    responses[response]["voting"] = self.__server_voting_time
                    responses[response]["now"] = Daytime.from_strftime(responses[response]["now"])

                    lo_now = Daytime.now()
                    tmp = abs(lo_now - responses[response]["now"])
                    if tmp > 0:
                        tmp = (3600*24) - tmp

                    self.__server_time_delta = Daytime.from_abs(tmp)

                    responses[response]["difference"] = self.__server_time_delta
                    responses[response]["until_voting"] = self.__server_voting_time - responses[response]["now"]

                case _:
                    if response.startswith("gRes"):  # because it could also be "gRes|last" or "gRes+now"
                        res = responses[response]
                        out = dict()

                        for voting in res:
                            attendants = dict()  # create dictionary with all attendants: votes
                            now_voting = res[voting]
                            everyone = [now_voting[element] for element in now_voting]
                            everyone += (['Lukas', 'Niclas', 'Melvin'] if voting == 'GayKing' else [])
                            for element in everyone:
                                attendants[element] = 0

                            votes = int()
                            for element in res[voting]:  # assign votes to attendant
                                votes += 1
                                attendants[res[voting][element]] += 1
                            out[voting] = dict()
                            out[voting]['totalVotes'] = votes
                            out[voting]['results'] = attendants

                        responses[response] = out

        return responses

    def _send(self, dictionary: dict, wait: bool = False) -> float | None:
        """
        send messages to server

        :param dictionary: dict to send
        :param wait: don't send the messages immediately and wait
        :return: time of sending
        """
        if not self:
            raise AuthError("Cannot send message - not authenticated yet!")
        self.__message_pool.append(dictionary)
        if not wait:
            return self.__send()

    def __send(self) -> float:
        """
        the actual sending process
        """
        # check errors before executing
        if not self.server_ip:
            raise Error("no host set")

        if not self.__bool__():
            raise AuthError("Not authenticated")

        if len(self.__message_pool) == 0:
            raise ValueError("Client: Message Pool Empty")

        for element in self.__message_pool:
            if "message" in element:
                element["message"] = element["message"].replace("'", "\'").replace('"', '\"')

        message = {
            "time": time.time(),
            "content": self.__message_pool
        }

        if self.__AuthKey:
            string_mes = json.dumps(message, ensure_ascii=True)
            mes = cryption_tools.MesCryp.encrypt(string_mes, key=self.__AuthKey.encode())

            try:
                self.Server.send(mes)
                # reset message pool only if the message was sent successfully
                self.__message_pool = []

            except BrokenPipeError:
                self.end(revive=True)
                return -1

            if self._debug_mode in ('normal', 'full'):
                print(ConsoleColors.OKCYAN+string_mes+ConsoleColors.ENDC)
            if self._debug_mode == 'full':
                print(ConsoleColors.WARNING+str(mes)+ConsoleColors.ENDC)

            return message["time"]

        string_mes = json.dumps(message, ensure_ascii=False)
        self.Server.send(cryption_tools.MesCryp.encrypt(string_mes))

        # reset message pool only if the message was sent successfully
        self.__message_pool = []
        if self._debug_mode in ('normal', 'full'):
            print(ConsoleColors.OKCYAN+string_mes+ConsoleColors.ENDC)

        return message["time"]

    def send(self) -> dict:
        """
        send the messages and also receive them
        """
        res = self.wait_for_message(self.__send())
        return res

    def __assign_results(self, results: dict) -> None:
        if self._debug_mode in ("normal", "full"):
            print(f"assigning {results=}")

        if results is ...:
            raise ValueError("results not set")

        for element in results.keys():
            # check if the element is in both list (using sets)
            if not set(results.keys()) & set(self.__results_getters.keys()) & {element}:
                raise ValueError(f"element {element} not in results and getters")

            if self._debug_mode == "full":
                print(f"appended {element=}, now: {self.results_getters}")
            self.__results_getters[element].result = results[element]

        for element in set(self.__results_getters.keys()) - set(results.keys()):
            self.__results_getters[element].result = False

    @property
    def results_getters(self) -> dict:
        return self.__results_getters

    @DEBUGGER.catch_and_write(raise_error=True, print_traceback=True)
    def receive(self):
        """
        receive messages from server, decrypt them and raise incoming errors (meant as thread, run by default)
        """
        while self.loop:
            try:
                bs = self.Server.recv(8)    # receive message length
                (length,) = struct.unpack('>Q', bs)

            except (ConnectionResetError, struct.error, socket.timeout):
                continue
            if self._debug_mode == "full":
                print(f"new message, receiving now")

            data = b''
            no_rec = 0
            to_read = 0
            while len(data) < length:   # receive message in patches so size doesn't matter
                o_to_read = to_read
                to_read = length - len(data)
                data += self.Server.recv(
                                    4096 if to_read > 4096 else to_read
                                    )

                if to_read == o_to_read:    # check if new packages were received
                    no_rec += 1
                else:
                    no_rec = 0

                if no_rec >= 100:          # if for 100 loops no packages were received, raise connection loss
                    raise socket.error('Failed receiving data - connection loss')

            if self._debug_mode == "full":
                print("received data")

            try:
                mes = cryption_tools.MesCryp.decrypt(data, self.__AuthKey.encode())
            except cryption_tools.InvalidToken:
                self._messages["Error"] = {"Error": "MessageError", "info": f"cant decrypt: {data}"}
                continue

            if self._debug_mode in ("full", "normal"):
                print(f"decrypted data: {mes=}")

            try:
                for _ in range(2):
                    mes = mes.replace("\\\\", "\\")
                mes = json.loads(mes)

            except json.decoder.JSONDecodeError:
                self._messages["Error"] = f"cant decode: {mes}, type: {type(mes)}"
                continue

            try:
                # parse each message
                for resp_type, message in mes["content"].items():
                    match message["type"]:
                        case "function":
                            if mes["time"] not in self._messages:
                                self._messages[mes["time"]] = {}

                            self._messages[mes["time"]][resp_type] = message["content"]

                        case "Error":
                            self._messages["Error"] = message["content"]

                        case "disconnect":
                            self._messages["disconnect"] = True
                            self.end()

                        case "ServerRequest":
                            self._server_messages[mes['time']] = message

                        case _:
                            raise ServerError(f"server send message: {mes}")

            except KeyError:
                with open("backend.err.log", 'a') as out:
                    out.write(format_exc()+f'message: {mes}')
                raise

    def wait_for_message(self, time_sent: float, timeout: int | bool | None = 10, delay: int | None = .1) -> dict:
        """
        wait for the server message to be received.
        :param time_sent: the time the message was sent
        :param timeout: raise an error if no correct message was received (seconds)
        :param delay: The delay for the while loop when checking self.messages
        :return: message(dict)
        """
        start = time.time()
        if self._debug_mode in ('full', 'normal'):
            print(f'waiting for message: {time_sent}')

        while time_sent not in self._messages:  # wait for server message
            if self._debug_mode == 'full':
                print(self._messages)

            if timeout and time.time()-start >= timeout:
                raise NetworkError("no message was received from server before timeout")

            elif "Error" in self._messages:
                for _ in range(10):
                    if time_sent in self._messages:
                        break
                    time.sleep(.01)
                break

            elif "disconnect" in self._messages:
                raise ConnectionAbortedError("Server ended connection")
            time.sleep(delay)

        with suppress(KeyError):
            out = self._messages[time_sent]
            del self._messages[time_sent]

            if len(out) == 0:
                raise MessageError("received empty message pool from server")

            out = self.response_handler(out)
            self.__assign_results(out)

        if "Error" in self._messages:
            try:
                error_name, full_error = self._messages["Error"]["Error"],  self._messages["Error"]

            except TypeError:
                raise Error(self._messages.pop("Error"))

            self._messages.pop("Error")
            self.error_handler(error_name, full_error)
            return {}

        if self._debug_mode in ('all', 'normal'):
            print(f"found message: {out}")

        return out

    def reconnect(self) -> None:
        """
        reconnect to server
        """
        try:    # try to reconnect to the server
            self.Server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.Server.connect((self.server_ip, self.port))  # connect to server
        except socket.error:
            raise ConnectionError('Server not reachable')

    # user functions
    def auth(self, username: str, password: str, pwd_hashed: bool = False) -> bool:
        """
        authenticate with the server
        """
        if not self.server_ip:
            raise Error("no host set")

        if not self.loop:
            raise Error("already called 'end'")

        if self:
            self.end(revive=True)

        self.reconnect()
        msg = {  # message
            'type': 'auth',
            'Name': username,
            'pwd': sha512(password.encode()).hexdigest() if not pwd_hashed else password,
            "com_protocol_version": COMM_PROTOCOL_VERSION
        }
        self.__userN = username
        self.__pwd_hash = msg["pwd"]
        self.__AuthKey = None  # reset AuthKey
        string_mes = json.dumps(msg, ensure_ascii=False)

        mes = cryption_tools.MesCryp.encrypt(string_mes)
        self.Server.send(mes)

        mes = json.loads(cryption_tools.MesCryp.decrypt(self.Server.recv(2048)))
        if "Error" in mes:
            self.error_handler(mes["Error"], mes)
        self.__AuthKey = mes['AuthKey']

        # setting timeout for receive thread
        self.Server.settimeout(.5)

        self.receive_thread = self.executor.submit(self.receive)  # start thread for receiving
        if mes["Auth"]:
            self.__login_time = Daytime.now()

        return mes['Auth']  # return True or False

    def re_auth(self) -> bool:
        """
        if a user has already been logged in, you can now try to re-login
        """
        if self.__userN is None or self.__pwd_hash is None:
            return False

        self.end(revive=True)
        return self.auth(username=self.__userN, password=self.__pwd_hash, pwd_hashed=True)

    def get_sec_clearance(self, wait: bool = False) -> str | FridrichFuture:
        """
        if signed in, get security clearance
        """
        msg = {'type': 'secReq'}

        self._send(msg, wait=True)

        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def send_vote(self, *args, flag: str | None = 'vote', voting: str | None = 'GayKing', wait: bool = False) -> None:
        """
        send vote to server\n
        flag can be "vote", "unvote", "dvote" or "dUvote", voting is custom\n
        DoubleVotes are only available once a week\n
        types will be ignored if flag is "dvote"
        """
        msg = {
               'type': flag,
               'voting': voting
        }
        if flag in ('vote', 'dvote'):
            msg['vote'] = args[0]  # if vote send vote

        self._send(msg, wait=True)

        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def get_results(self, flag: str | None = 'now', wait: bool = False) -> dict | FridrichFuture:
        """
        get results of voting\n
        flag can be "now", "last"\n
        return format: {voting : {"totalvotes" : int, "results" : {name1 : votes, name2 : votes}}}
        """
        msg = {
               'type': 'gRes',
               'flag': flag,
               'f_name': "gRes"+flag
        }    # set message
        self._send(msg, wait=True)

        res = FridrichFuture()
        self.__results_getters[msg["f_name"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_log(self, wait: bool = False) -> dict | FridrichFuture:
        """
        get list of recent GayKings
        """
        msg = {
               'type': 'gLog'
        }   # set message
        self._send(msg, wait=True)

        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    @staticmethod
    def calculate_streak(log: dict) -> tuple[str, int]:
        """
        if someone got voted multiple times in a row,
        return their name and how often they
        got voted
        :return: (Name, Streak)
        """
        # sort list by year, month, date
        sorted_log = {x: log[x] for x in sorted(log, key=lambda x: '.'.join(reversed(x.split('.'))))}
        try:
            streak_guys: dict = {guy: 0 for guy in sorted_log[list(sorted_log.keys())[-1]].split("|")}
        except AttributeError:
            raise ValueError("Please only pass one voting per time")

        for guy in streak_guys.keys():
            for date in reversed(sorted_log.keys()):
                if guy in sorted_log[date] and streak_guys[guy] != -1:
                    streak_guys[guy] += 1
                    continue
                break

        max_num = max(list(streak_guys.values()))
        names = "|".join([guy for guy in streak_guys if streak_guys[guy] == max_num])
        return names, max_num  # return results

    def get_cal(self, wait: bool = False) -> dict | FridrichFuture:
        """
        get Calendar in format {"date":listOfEvents}
        """
        msg = {
               'type': 'gCal'
        }   # set message
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def send_cal(self, date: str, event: str, wait: bool = False) -> None:
        """
        send entry to calender
        """
        msg = {
               'type': 'CalEntry',
               'date': date,
               'event': event
        }   # set message
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def change_pwd(self, new_password: str, wait: bool = False) -> None:
        """
        Change password of user currently logged in to
        """
        msg = {
               'type': 'changePwd',
               'newPwd': sha512(new_password.encode()).hexdigest()
        }    # set message
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def get_vote(self, flag: str | None = 'normal', voting: str | None = 'GayKing',
                 wait: bool = False) -> str | FridrichFuture:
        """
        get current vote of user\n
        flag can be normal or double
        """
        msg = {
               'type': 'getVote',
               'flag': flag,
               'voting': voting,
               'f_name': "getVote"+flag
        }    # set message
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["f_name"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_version(self, wait: bool = False) -> str | FridrichFuture:
        """
        get current version of GUI program
        """
        msg = {
               'type': 'getVersion'
        }  # set message
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def set_version(self, version: str, wait: bool = False) -> None:
        """
        set current version of GUI program
        """
        msg = {
               'type': 'setVersion',
               'version': version
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def get_frees(self, wait: bool = False) -> int | FridrichFuture:
        """
        get free double votes
        """
        msg = {
               'type': 'getFrees'
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_online_users(self, wait: bool = False) -> list | FridrichFuture:
        """
        get list of currently online users
        """
        msg = {
               'type': 'gOuser'
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def send_chat(self, message: str, wait: bool = False) -> None:
        """
        send message to chat
        """
        msg = {
               'type': 'appendChat',
               'message': message
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def get_chat(self, wait: bool = False) -> list | FridrichFuture:
        """
        get list of all chat messages
        """
        msg = {
               'type': 'getChat'
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_weather_stations(self, wait: bool = False) -> List[Dict[str, str]] | FridrichFuture:
        """
        get the names and locations of all registered weather stations
        """
        msg = {
            "type": "get_stations"
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_temps_now(self, wait: bool = False) -> dict | FridrichFuture:
        msg = {
            "type": "get_temps_now"
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_temps_log(self, station_name: str, wait: bool = False) -> dict | FridrichFuture:
        msg = {
            "type": "get_temps_log",
            "station_name": station_name
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    # user controlled variables:
    def get_all_vars(self, wait: bool = False) -> dict | FridrichFuture:
        """
        get all user controlled variables inside a dict
        """
        msg = {
            "type": "get_all_vars"
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_var(self, variable: str, wait: bool = False) -> Any | FridrichFuture:
        """
        get a user controlled variable
        """
        msg = {
            "type": "get_var",
            "var": variable
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def set_var(self, variable: str, value, wait: bool = False) -> None:
        """
        set a user controlled variable
        must be json valid!
        """
        msg = {
            "type": "set_var",
            "var": variable,
            "value": value
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def del_var(self, variable: str, wait: bool = False) -> None:
        """
        delete a user controlled variable
        """
        msg = {
            "type": "del_var",
            "var": variable
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def __iter__(self) -> Iterable:
        """
        return dict of all User Controlled Variables when called
        """
        _d = self.get_all_vars()
        for element in _d:
            yield element, _d[element]

    def __getitem__(self, item: str):
        return self.get_var(item)

    def __setitem__(self, key: str, value) -> None:
        return self.set_var(key, value)

    def __delitem__(self, item: str) -> None:
        return self.del_var(item)

    # WeatherStation Funcs
    def register_station(self, station_name: str, location: str, wait: bool = False) -> Union[FridrichFuture, None]:
        """
        register a new weather station
        """
        msg = {
            "type": "register",
            "station_name": station_name,
            "location": location
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def commit_weather_data(self, station_name: str, weather_data: dict,
                            wait: bool = False, set_time: str | None = ...) -> Union[None, FridrichFuture]:
        """
        Commit data to the WeatherStation database
        """
        msg = {
            "type": "commit",
            "time": time.strftime("%Y.%m.%d")+"-"+Daytime.now().to_string() if set_time is ... else set_time,
            "station_name": station_name
        }
        msg.update(weather_data)
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    # Admin Functions
    def admin_get_users(self, wait: bool = False) -> list | FridrichFuture:
        """
        get list of all users with passwords and security clearance\n
        return format: [{"Name":username, "pwd":password, "sec":clearance}, ...]
        """
        msg = {
               'type': 'getUsers'
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def admin_set_password(self, user: str, password: str, wait: bool = False) -> None:
        """
        set password of given user
        """
        msg = {
               'type': 'setPwd',
               'User': user,
               'newPwd': sha512(password.encode()).hexdigest()
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def admin_set_username(self, old_username: str, new_username: str, wait: bool = False) -> None:
        """
        change username of given user
        """
        msg = {
               'type': 'setName',
               'OldUser': old_username,
               'NewUser': new_username
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def admin_set_security(self, username: str, password: str, wait: bool = False) -> None:
        """
        change security clearance of given user
        """
        msg = {
               'type': 'setSec',
               'Name': username,
               'sec': password
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def admin_add_user(self, username: str, password: str, clearance: str, wait: bool = False) -> None:
        """
        add new user
        """
        msg = {
               'type': 'newUser',
               'Name': username,
               'pwd': sha512(password.encode()).hexdigest(),
               'sec': clearance
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def admin_remove_user(self, username: str, wait: bool = False) -> None:
        """
        remove user
        """
        msg = {
               'type': 'removeUser',
               'Name': username
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def admin_reset_logins(self, wait: bool = False) -> None:
        """
        reset all current logins
        """
        msg = {
               'type': 'rsLogins'
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def manual_voting(self, wait: bool = False) -> None:
        """
        trigger a voting manually
        """
        msg = {
            "type": "trigger_voting"
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    def kick_user(self, user_to_kick: str, wait: bool = False) -> None:
        """
        kick one user by its username
        """
        msg = {
            "type": "kick_user",
            "user": user_to_kick
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()

    # AppStore functions
    def get_apps(self, wait: bool = False) -> list | FridrichFuture:
        """
        get all available apps and versions
        """
        msg = {
            "type": "get_apps"
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def download_app(self, app: str, directory: str | None = ...) -> None:
        """
        :param app: which app to download
        :param directory: where the program should be downloaded to
        """
        msg = {
            "type": "download_app",
            "app": app
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        self.send()
        meta = res.result

        self.load_state = "Uploading"
        for _ in meta:
            thread = send_receive(mode='receive', print_steps=False, download_directory=directory, thread=True,
                                  overwrite=True)
            while thread.running():
                pass
                self.load_program = now_download_program
                self.load_progress = now_download_progress
        self.load_state = str()
        self.load_program = str()
        self.load_progress = float()

    def _send_app(self, files: list | tuple, app_name: str) -> None:
        for file in files:
            thread = send_receive(mode="send", filename=file, destination=self.server_ip, print_steps=False,
                                  thread=True, overwrite=True)
            while thread.running():
                self.load_program = app_name
        self.load_program = str()
        self.load_state = str()

    def create_app(self, app_name: str, app_version: str, app_info: str, files: list | tuple) -> None:
        """
        add a new app to the fridrich appstore
        """
        msg = {
            "type": "create_app",
            "name": app_name,
            "version": app_version,
            "info": app_info,
            "files": [file.split("/")[-1] for file in files]
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        self.send()

        self.load_state = "Uploading"
        self._send_app(files, app_name)

    def modify_app(self, old_app_name: str, app_name: str, app_version: str, app_info: str,
                   files: list | tuple, to_delete: list | tuple) -> None:
        """
        configure an already existing app

        :param old_app_name: the original name of the app
        :param app_name: the name of the app
        :param app_version: the version of the app
        :param app_info: the info of the app
        :param files: a list with files to update (full path, overwriting old files)
        :param to_delete: a list with App-Files that should be deleted (if the app exist!)
        """
        msg = {
            "type": "modify_app",
            "o_name": old_app_name,
            "name": app_name,
            "version": app_version,
            "info": app_info,
            "files": [file.split("/")[-1].split("\\")[-1] for file in files],
            "to_remove": to_delete
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        self.send()

        self._send_app(files, app_name)

    # tools
    def ping(self, wait: bool = False) -> float | FridrichFuture:
        """
        ping the server to check the connection time
        :return: time in ms
        """
        msg = {
            "type": "ping",
            "time": time.time()
        }
        self._send(msg, wait=True)

        # result handling
        res = FridrichFuture()
        self.__results_getters[msg["type"]] = res
        if not wait:
            self.send()
            return res.result
        return res

    def get_server_time(self, wait: bool = False, reload: bool = False) -> dict | FridrichFuture:
        """
        get the current server time and voting time
        :param reload: if already called once and False, doesn't request new time from server (local update)
        :param wait: if reload is set, sets if waits for Connection.send()
        """
        res = FridrichFuture()
        if not self.__server_time_delta or reload:
            msg = {
                "type": "get_time"
            }
            self._send(msg, wait=True)

            # result handling
            self.__results_getters[msg["type"]] = res
            if not wait:
                self.send()
                return res.result
            return res

        tmp = Daytime.now() + self.__server_time_delta
        res.result = {
            "now": tmp,
            "voting": self.__server_voting_time,
            "difference": self.__server_time_delta,
            "until_voting": self.__server_voting_time - tmp
        }
        if not wait:
            return res.result
        return res

    # magical functions
    def __repr__(self) -> str:
        return f'<Backend instance (debug_mode: {self._debug_mode}, user: {self.__userN})>'

    def __str__(self) -> str:
        """
        return string of information when str() is called
        """
        return self.__repr__()

    def __bool__(self) -> bool:
        """
        return True if AuthKey
        """
        return bool(self.__AuthKey)

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, exception_type, value, traceback) -> bool:
        self.end(revive=False)
        if exception_type is not None:
            return False
        return True

    def __del__(self) -> None:
        self.end(revive=False)

    def __eq__(self, other: "Connection") -> bool:
        if not type(other) == Connection:
            return self.__bool__()

        return all([str(self) == str(other), bool(self) is bool(other), self.server_ip == other.server_ip,
                    self.port == other.port])

    # the end
    def end(self, revive: bool = False) -> None:
        """
        close connection with server and logout
        """
        msg = {
               'type': 'end'
        }    # set message
        with suppress(ConnectionResetError, ConnectionAbortedError, AuthError):
            self._send(msg, wait=False)  # send message

        self.loop = False
        self.__AuthKey = None
        self.__login_time = None

        if revive:
            # waiting for receive thread to shut down
            while not self.receive_thread.done():
                time.sleep(.1)

            # re-creating the thread-pool
            del self.executor
            self.executor = ThreadPoolExecutor(max_workers=1)
            self.loop = True
            return

        self.executor.shutdown(wait=True)
        # app_store.executor.shutdown(wait=False)


############################################################################
#                             helper functions                             #
############################################################################
def date_for_sort(message: dict) -> str:
    """
    go from format
    "hour:minute:second:millisecond - day.month.year" to "year.month.day - hour:minute:second:millisecond"
    """
    y = message['time'].split(' - ')    # split date and time
    return '.'.join(reversed(y[1].split('.')))+' - '+y[0]   # reverse date and place time at end



"""
helper classes used by server and backend

Author:
Nilusink
"""
import typing
import json
import time


class Daytime:
    """
    class for calculating with HH:MM:SS
    """
    def __init__(self, hour: int = 0, minute: int = 0, second: int = 0) -> None:
        self.hour = hour
        self.minute = minute
        self.second = second

    @property
    def hour(self) -> int:
        return self.__hour

    @hour.setter
    def hour(self, value: int) -> None:
        self.__hour = value % 24
        while self.__hour < 0:
            self.__hour += 24

    @property
    def minute(self) -> int:
        return self.__minute

    @minute.setter
    def minute(self, value: int) -> None:
        self.__minute = value % 60
        while self.__minute < 0:
            self.__minute += 60

    @property
    def second(self) -> int:
        return self.__second

    @second.setter
    def second(self, value: int) -> None:
        self.__second = value % 60
        while self.__second < 0:
            self.__second += 60

    @staticmethod
    def from_abs(absolute_value: int) -> "Daytime":
        """
        convert an absolute value (seconds) to a Daytime

        :param absolute_value: the absolute time value in seconds
        """
        if type(absolute_value) != int:
            raise ValueError("only accepts of type int")

        # making sure the value is in range of the 24 hour time format
        while absolute_value < 0:
            absolute_value += 24*3600

        while absolute_value > 24*3600:
            absolute_value -= 24*3600

        # splitting up in Hour, Minute, Second
        hour = absolute_value // 3600
        absolute_value -= hour * 3600
        minute = absolute_value // 60
        absolute_value -= minute * 60
        return Daytime(hour=hour, minute=minute, second=absolute_value)

    @staticmethod
    def from_strftime(string_time: str, sep: str = ":") -> "Daytime":
        """
        from format "HH:MM:SS"
        """
        parts = string_time.split(sep)
        parts += ["00"] * (3-len(parts))
        return Daytime(hour=int(parts[0]), minute=int(parts[1]), second=int(parts[2]))

    @staticmethod
    def now() -> "Daytime":
        """
        Daytime object with current time
        """
        return Daytime(hour=int(time.strftime("%H")), minute=int(time.strftime("%M")), second=int(time.strftime("%S")))

    # mathematics
    def __add__(self, other: "Daytime") -> "Daytime":
        return Daytime.from_abs(abs(self) + abs(other))

    def __iadd__(self, other: "Daytime") -> "Daytime":
        self.__dict__ = self.__add__(other).__dict__
        return self

    def __sub__(self, other: "Daytime") -> "Daytime":
        return Daytime.from_abs(abs(self) - abs(other))

    def __isub__(self, other: "Daytime") -> "Daytime":
        self.__dict__ = self.__sub__(other).__dict__
        return self

    # comparison
    def __eq__(self, other: "Daytime") -> bool:
        return abs(self) == abs(other)

    def __ne__(self, other: "Daytime") -> bool:
        return abs(self) != abs(other)

    def __lt__(self, other: "Daytime") -> bool:
        return abs(self) < abs(other)

    def __le__(self, other: "Daytime") -> bool:
        return abs(self) <= abs(other)

    def __gt__(self, other: "Daytime") -> bool:
        return abs(self) > abs(other)

    def __ge__(self, other: "Daytime") -> bool:
        return abs(self) >= abs(other)

    # representation
    def __str__(self) -> str:
        return f"{'0' if self.hour < 10 else ''}{self.hour}:{'0' if self.minute < 10 else ''}" \
               f"{self.__minute}:{'0' if self.second < 10 else ''}{self.__second}"

    def to_string(self) -> str:
        """
        call __str__
        """
        return self.__str__()

    def __repr__(self) -> str:
        return f"<Daytime: {self.__str__()}>"

    def __bool__(self) -> bool:
        return bool(self.__abs__())

    def __abs__(self) -> int:
        """
        returns time in seconds
        """
        return self.hour * 3600 + self.minute * 60 + self.second

    # accessibility
    def __getitem__(self, item: str) -> int:
        match item:
            case "hour":
                return self.hour
            case "minute":
                return self.minute
            case "second":
                return self.second
            case _:
                raise ValueError(f"{item} is not a valid variable!")

    def __setitem__(self, item: str, value: int) -> None:
        match item:
            case "hour":
                self.hour = value
            case "minute":
                self.minute = value
            case "second":
                self.second = value
            case _:
                raise ValueError(f"{item} is not a valid variable!")

    def __delitem__(self, item: str) -> None:
        match item:
            case "hour":
                self.hour = 0
            case "minute":
                self.minute = 0
            case "second":
                self.second = 0
            case _:
                raise ValueError(f"{item} is not a valid variable!")


class FileVar:
    def __init__(self, value: str | dict, files: str | list | tuple) -> None:
        """
        create a variable synced to one or more files
        """
        # filter all items that are not a string
        self.files = [file if type(file) == str else ... for file in files] if type(files) in (list, tuple) else [files]
        # remove all non standard items
        while ... in self.files:
            self.files.remove(...)

        self.value = value
        self.type = type(value)

        self.set(value)  # assign variable

    def __repr__(self) -> str:
        self.get()
        return repr(self.value)

    def __len__(self) -> int:
        """
        return the length of ``self.value``
        """
        self.get()
        return len(self.value)

    def __str__(self) -> str:
        """
        return string of ``self.value``
        """
        self.get()
        return str(self.value)

    # str options
    def __add__(self, other: str) -> str:
        self.get()
        self.check_type(str)

        self.set(self.value + other)

        return self.value

    # dict options
    def __getitem__(self, key: str) -> typing.Any:
        """get an item if ``self.value`` is a dict
        """
        self.get()  # update variable in case something in the file has changed
        self.check_type(dict)
        if key not in self.value:
            raise KeyError(f'"{key}" not in dict "{self.value}"')
        return self.value[key]

    def __setitem__(self, key, value) -> dict:
        """
        set an item of a dict
        """
        self.get()
        self.check_type(dict)
        self.value[key] = value
        self.set(self.value)
        return self.value

    def __delitem__(self, key: str) -> None:
        self.get()
        self.check_type(dict)

        del self.value[key]

    def __iter__(self) -> typing.Iterator:
        self.get()
        self.check_type(dict)

        for key, item in self.value.items():
            yield key, item

    # general options
    def __eq__(self, other) -> bool:
        """
        check if the ``==`` given value is the same as either the whole class or the value
        """
        self.get()
        if type(other) == FileVar:
            return other.value == self.value and list(other.files) == list(self.files)

        return other == self.value

    def __contains__(self, other) -> bool:
        self.get()
        return other in self.value

    def set(self, value: str | dict) -> None:
        """
        set the variable (update files)
        """
        self.value = value
        self.type = type(value)

        for file in self.files:
            for _ in range(5):  # check five times if a file is already opened, if failed then pass
                try:
                    with open(file, 'w') as out:
                        if self.type == dict:
                            json.dump(self.value, out, indent=4)
                            break
                        out.write(self.value)
                        break

                except IOError:
                    time.sleep(.2)

                except json.JSONDecodeError:
                    with open(file, "w") as out_p:
                        json.dump({}, out_p)

            else:
                raise IOError(f"Can't access file {file}")

    def get(self) -> str | dict:
        """
        get the variable in its original type
        """
        file = self.files[0]
        for _ in range(5):
            try:
                with open(file, 'r') as inp:
                    try:
                        self.value = json.load(inp)

                    except json.JSONDecodeError:
                        self.value = inp.read()

                self.type = type(self.value)

                return self.value

            except IOError:
                time.sleep(.2)

            except json.JSONDecodeError:
                with open(file, "w") as out_p:
                    json.dump({}, out_p)

        else:
            raise IOError(f"Can't access file {file}")

    def check_type(self, wanted_type: typing.Type[str] | typing.Type[dict]) -> None:
        """
        if type is wrong, raise an error
        """
        if not self.type == wanted_type:
            raise TypeError(
                f'Expected {wanted_type}, got {self.type}. This function is not available for the given variable type')



"""
server sub-module
Contains all the modules that only the server needs

Author: Nilusink
"""
from fridrich import ConsoleColors
from dataclasses import dataclass
import traceback
import datetime
import typing
import types
import json
import os


@dataclass(init=False)
class Constants:
    """
    All constants (modify in file settings.json)
    """
    port: int | None = ...
    ip: str | None = ...
    Terminate: bool | None = ...

    direc: str | None = ...

    lastFile: str | None = ...
    nowFile: str | None = ...
    strikeFile: str | None = ...

    logDirec: str | None = ...

    CalFile: str | None = ...
    crypFile: str | None = ...
    versFile: str | None = ...
    tempLog: str | None = ...
    doubFile: str | None = ...
    SerlogFile: str | None = ...
    SerUpLogFile: str | None = ...
    ChatFile: str | None = ...
    VarsFile: str | None = ...
    WeatherDir: str | None = ...

    logFile: str | None = ...
    errFile: str | None = ...
    tempFile: str | None = ...

    DoubleVotes: int | None = ...

    DoubleVoteResetDay: str | None = ...
    switchTime: str | None = ...
    rebootTime: str | None = ...
    status_led_pin: int | None = ...
    status_led_sleep_time: list | None = ...

    AppStoreDirectory: str | None = ...

    def __init__(self) -> None:
        """
        create instance
        """
        # get variable values
        try:
            self.dic = json.load(open(os.getcwd() + '/config/settings.json', 'r'))

        except FileNotFoundError:
            self.dic = json.load(open('/home/apps/Fridrich/config/settings.json', 'r'))

        for Index, Value in self.dic.items():
            setattr(self, Index, Value)

    def __getitem__(self, item) -> str | int | bool:
        return self.dic[item]


class Debug:
    """
    for debugging...
    """
    instance: "Debug" = None

    # if a instance already exists, pass that one
    def __new__(cls, *args, **kw):
        if cls.instance is not None:
            print("old instance!")
            return cls.instance

        cls.instance = super(Debug, cls).__new__(cls)
        return cls.instance

    def __init__(self, deb_file: str, error_file: str) -> None:
        """
        debFile: file to write debug-messages to
        """
        self.file = deb_file
        self.errFile = error_file

        with open(self.file, 'w') as out:
            out.write('')

        with open(self.errFile, 'a') as out:
            out.write(
                f'\n\n\n\n\n######## - Program restart [{datetime.datetime.now().strftime("%Y.%m.%d at %H:%M:%S.%f")}] - ########\n\n')

    def debug(self, *args) -> None:
        """
        prints and writes all arguments

        for each argument a new line in the file is begun
        """
        print(*args)
        with open(self.file, 'a') as out:
            for element in args:
                out.write(str(element) + '\n')

    def catch_traceback(self, raise_error: bool = False) -> typing.Callable:
        """
        execute function with traceback and debug all errors
        """
        def decorator(func: types.FunctionType) -> typing.Callable:
            def wrapper(*args, **kw) -> None:
                try:
                    return func(*args, **kw)

                except Exception as e:
                    err = f'{ConsoleColors.FAIL}\n\n\n######## - Exception "{e}" at function {func.__name__} on {datetime.datetime.now().strftime("%H:%M:%S.%f")} -' \
                          f' ########\n\n{traceback.format_exc()}\n\n######## - END OF EXCEPTION - ########\n\n\n{ConsoleColors.ENDC}'
                    self.debug(err)
                    if raise_error:
                        raise

            return wrapper
        return decorator

    def write_traceback(self, error: type, from_user: str | None = ...) -> None:
        """
        write a caught error
        """
        err = '\n\n\n' + (
            "From User: " + from_user if from_user is not ... else "") + f'######## - Exception "{error}" on {datetime.datetime.now().strftime("%H:%M:%S.%f")} - ########\n\n{traceback.format_exc()}\n\n######## - END OF EXCEPTION - ########\n\n\n'
        self.debug(err)


Const = Constants()
DEBUGGER = Debug(Const.SerlogFile, Const.errFile)



"""
Debugging helper for backend

Author:
Nilusink
"""
from typing import Callable, Tuple, Any
from fridrich import ConsoleColors
from traceback import format_exc
import time
import os


class Debugger:
    """
    functions for advanced debugging
    """
    total_title_length: int = 60

    def __init__(self, outfile: str, file_mode: str = "a") -> None:
        """
        :param outfile: the file to save the error logs to
        :param file_mode: should be "w" or "a" (FileIO write mode)
        """
        self.__outfile = outfile
        self.__file_mode = file_mode

        # check if the directory to save in already exists
        direc = os.path.split(self.__outfile)[0]
        if not os.path.isdir(direc):
            os.mkdir(direc)

        with open(self.__outfile, file_mode) as out:
            date = time.strftime("%Y.%m.%d - %H:%M:%S")
            side1, side2 = self.__calculate_sides(self.total_title_length, len(date))

            # create "heading" every time the program restarts
            out.write(f"\n\n\n\n{'#'*self.total_title_length}\n#{' '*side1}{date}"
                      f"{' '*side2}#\n{'#'*self.total_title_length}")

        print(f"File for debugging: \"{self.__outfile}\"")

    @staticmethod
    def __calculate_sides(total_length: int, string_length: int, spaces: int | None = 2) -> Tuple[int, int]:
        """
        used for creating more readable errors

        :param total_length: the total length of the finished string
        :param string_length: the length of the string to insert
        :param spaces: the amount of spaces planned between "#" and the inserted string
        """
        if total_length - string_length - spaces % 2 == 0:
            s1 = s2 = (total_length - string_length - spaces) // 2

        else:
            s1 = (total_length - string_length - spaces) // 2
            s2 = s1 + 1

        return s1, s2

    def catch_and_write(self, raise_error: bool = False, print_traceback: bool = False) -> Callable:
        """
        (decorator)
        catch tracebacks and write them into a file

        :param raise_error: defines if the caught error should be raised (still write to file and print if enabled)
        :param print_traceback: defines if the caught error should be printed
        """
        def decorator(function: Callable) -> Callable:
            def wrapper(*args, **kwargs) -> Any:
                try:
                    return function(*args, **kwargs)

                except (Exception,):
                    title = f"ERROR IN FUNCTION \"{function.__name__}\""
                    end = f"END OF EXCEPTION"
                    h1, h2 = self.__calculate_sides(self.total_title_length, len(title))
                    e1, e2 = self.__calculate_sides(self.total_title_length, len(end))

                    trace = f"\n\n\n{h1*'#'} {title} {h2*'#'}\n\n{format_exc()}\n\n{e1*'#'} {end} {e2*'#'}\n\n\n"
                    try:
                        with open(self.__outfile, self.__file_mode) as out:
                            out.write(trace)

                        if print_traceback:
                            print(f"{ConsoleColors.FAIL}{trace}{ConsoleColors.ENDC}")

                    except NameError:
                        return

                    if raise_error:
                        raise
                    print("caught traceback")

            return wrapper
        return decorator



"""
for sending and receiving files from a remote host.

Author:
Nilusink
"""
from concurrent.futures import ThreadPoolExecutor, Future
import socket
import struct
import time
import json
import os

download_progress: float = 0.0
download_program: str = ""

executor = ThreadPoolExecutor()


def send_receive(mode: str, filename: str | None = ..., destination: str | None = ..., print_steps: bool | None = False,
                 download_directory: str | None = ..., thread: bool | None = False, overwrite: bool | None = False
                 ) -> None | Future:
    """
    send and receive files (function version)

    :param mode: either 's' | 'send' or 'r' | 'receive'
    :param filename: filename for sending files
    :param destination: ip/hostname of destination computer
    :param print_steps: enables print function when receiving
    :param download_directory: where the downloaded files should end up
    :param thread: if the program should be executed as a thread or not
    :param overwrite: if true overwrites output file
    :return: None of Future instance of the Thread
    """
    global download_progress, download_program
    if thread:
        return executor.submit(send_receive, mode=mode, filename=filename, destination=destination,
                               print_steps=print_steps, download_directory=download_directory, thread=False,
                               overwrite=overwrite)

    if mode in ('r', 'receive'):
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind(('0.0.0.0', 15151))
        server.listen()

        client, address = server.accept()
        resp = json.loads(client.recv(1024).decode('utf-8'))
        download_program = resp["filename"]
        client.send('received'.encode('utf-8'))

        # receiving data
        if resp['type'] == "file":
            bs = client.recv(8)
            (length,) = struct.unpack('>Q', bs)
            data = b''
            no_rec = 0
            to_read = 0
            start = time.time()
            while len(data) < length:
                # doing it in batches is generally better than trying
                # to do it all in one go, so I believe.
                o_to_read = to_read
                to_read = length - len(data)
                data += client.recv(
                                    4096 if to_read > 4096 else to_read
                                    )

                download_progress = len(data)/length

                if to_read == o_to_read:    # check if new packages were received
                    no_rec += 1
                else:
                    no_rec = 0

                if no_rec >= 100:          # if for 100 loops no packages were received, raise connection loss
                    raise socket.error('Failed receiving data - connection loss')

                if print_steps:
                    print(f'\rreceiving [{len(data)}/{length}]')

            if print_steps:
                print(f'receiving took {time.time()-start} sec.')

            filename = resp['filename']

            i = 0
            if not overwrite:
                print("not overwriting")
                while os.path.isfile(filename):  # check if file with the same name already exists
                    i += 1
                    parts = filename.split('.')
                    filename = (parts[0].rstrip(str(i-1))+str(i)+'.')+'.'.join(parts[1::])

                if filename != resp['filename']:
                    print(f'renamed file from "{resp["filename"]}" to "{filename}"')

            if download_directory is not ...:
                filename = download_directory+'/'+filename

            with open(filename, 'wb') as out:
                out.write(data)
            client.send("done".encode())
        else:
            print(f'Cannot receive of type "{resp["type"]}"')

    elif mode in ('s', 'send'):
        file_content = open(filename, 'rb').read()

        length = struct.pack('>Q', len(file_content))

        msg = {
            "type": "file",
            "filename": filename.split('/')[-1]
        }

        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.connect((destination, 15151))  # connect to server

        server.sendall(json.dumps(msg).encode('utf-8'))
        server.recv(1024)
        server.sendall(length)
        server.sendall(file_content)
        resp = str()
        while resp != "done":
            resp = server.recv(1024).decode()
            if resp != "done":
                print(f"invalid response: {resp}")

    else:
        raise ValueError(f"invalid parameter 'mode' with value '{mode}'")



from fridrich.file_transfer import send_receive
from fridrich.server.classes import User
import json
import os


def get_list() -> list:
    """
    :return: a list of available apps with versions
    """
    with open("/home/pi/Server/fridrich/server/settings.json", 'r') as inp:
        directory = json.load(inp)["AppStoreDirectory"]

    apps = list()
    for app in [dire for dire in os.listdir(directory) if os.path.isdir(directory + '/' + dire)]:
        size = float()
        filenames = [file for file in os.listdir(directory+app) if file.endswith(".zip")]
        if "AppInfo.json" not in os.listdir(directory+app):
            continue

        for filename in filenames:
            size += os.path.getsize(directory+app+'/'+filename)

        app_info = json.load(open(directory+app+'/AppInfo.json'))
        app_info["name"] = app
        app_info["files"] = filenames
        app_info["size"] = size
        apps.append(app_info)

    return apps


def send_apps(_message: dict, user: User) -> None:
    """
    :param _message: the message received from the client (not used)
    :param user: the user to send the answer to
    :return: None
    """
    user.send(get_list())


def download_app(message: dict, user: User) -> None:
    """
    :param message: the message received from the client (for the timestamp)
    :param user: the user to send the answer to
    :return: None
    """
    with open("/home/pi/Server/fridrich/server/settings.json", 'r') as inp:
        directory = json.load(inp)["AppStoreDirectory"]

    files = tuple((file for file in os.listdir(directory+message["app"]) if file.endswith(".zip")))

    user.send(files)
    for file in files:
        send_receive(mode="send", filename=directory+message["app"]+'/'+file, destination=user.ip, print_steps=False)


def receive_app(message: dict, user: User, modify: bool | None = False) -> None:
    """
    :param message: the message received from the client (for the timestamp)
    :param user: the user to send the answer to
    :param modify: if true used for modifying apps
    :return: None
    """
    with open("/home/pi/Server/fridrich/server/settings.json", 'r') as inp:
        directory = json.load(inp)["AppStoreDirectory"]+message["name"]

    if not modify:
        if os.path.isdir(directory):
            files = os.listdir(directory)
            if "AppInfo.json" in files:  # check if directory is empty, else send error
                user.send({"error": "ValueError", "info": f"App with name {message['name']} already exists"})
                return
            else:
                if len(files) != 0:
                    for element in files:
                        os.remove(directory+'/'+element)
        else:
            os.system(f"mkdir {directory}")

        with open(directory+"/AppInfo.json", 'w') as out:
            json.dump({
                       "version": message["version"],
                       "info": message["info"],
                       "publisher": user.name,
                       "publisher_id": user.id
            }, out, indent=4)

    user.send({"success": True})

    for _ in message["files"]:
        send_receive(mode='receive', print_steps=False, download_directory=directory, thread=False, overwrite=True)


def modify_app(message: dict, user: User) -> None:
    """
            msg = {
            "type": "modify_app",
            "o_name": old_app_name,
            "name": app_name,
            "version": app_version,
            "info": app_info,
            "new": [file.split("/").split("\\")[-1] for file in files],
            "to_delete": to_delete
    """
    try:
        app = {app["name"]: app for app in get_list()}[message["o_name"]]
    except KeyError:
        user.send({
                "error": f"App {message['o_name']} doesn't exist"
            })
        return

    if app["publisher_id"] != user.id:
        user.send({
                "error": f"App can only be modified by creator! {app['publisher']}"
            })
        return

    if app["publisher"] != user.name:
        app["publisher"] = user.name

    with open("/home/pi/Server/fridrich/server/settings.json", 'r') as inp:
        directory = json.load(inp)["AppStoreDirectory"]

    with open(directory+"/"+app["name"]+"/AppInfo.json", 'w') as out:
        print(f"before: {app}")
        tmp = {
            "version": "nAn",
            "info": message["info"],
            "publisher": user.name,
            "publisher_id": app["publisher_id"]
        }
        print(f"after: {tmp}")
        json.dump(tmp, out, indent=4)

    for file in message["to_remove"]:
        os.remove(directory+'/'+app['name']+'/'+file)

    if message["name"] != app["name"]:
        os.system(f"mv {directory+'/'+app['name']+'/'} {directory+'/'+message['name']+'/'}")

    receive_app(message, user, modify=True)

    with open(directory+"/"+app["name"]+"/AppInfo.json", 'w') as out:
        tmp = {
            "version": message["version"],
            "info": message["info"],
            "publisher": user.name
        }
        json.dump(tmp, out, indent=4)



"""
defines functions for the Server
(Server)

Author: Nilusink
"""
from fridrich.server.classes import *
from fridrich.classes import FileVar
from fridrich.server import Const
from sys import platform
from os import system
import contextlib
import datetime
import typing
import socket
import json
import time


def send_success(user: User) -> None:
    """
    send the success message to the client
    """
    user.send({'Success': 'Done'})


def check_if(s: str, d: dict, voting: str) -> str:
    """
    if the name is already in the dict, return the name in the dict

    else return the given name ("s")
    """
    if voting in d:
        d = d[voting]
        keys = [d[key] for key in d] + ['Lukas', 'Melvin',
                                        'Niclas']  # keys is (ex.) ['Fridrich', 'Lukas', 'Melvin', 'Niclas]

        for element in keys:
            if s.lower().replace(' ', '') == element.lower().replace(' ', ''):
                return element
        return s
    return s


def get_new_ones(flag: str, vote_instance: FileVar, last_file: str, voting: str) -> list:
    """
    get all attendants which are not in the default name list
    """
    new_ones = list()
    tmp: dict
    match flag:
        case 'now':
            tmp = vote_instance.get()
        case 'last':
            tmp = json.load(open(last_file, 'r'))
        case _:
            raise ValueError(f'"{flag}" is not an option')

    for element in tmp[voting]:
        if not tmp[voting][element] in ['Lukas', 'Niclas', 'Melvin'] + new_ones:
            new_ones.append(tmp[voting][element])

    return new_ones


def reboot() -> None:
    if platform == "win32":
        system("shutdown -r -t 0")
        return
    system("sudo reboot")


class Chat:
    """
    Handler for Chat file
    """

    @staticmethod
    def add(message: str, from_user: str) -> None:
        """
        append a message to the file
        """
        mes = Chat.get()  # get message list from file

        curr_time = datetime.datetime.now()
        formatted_time = curr_time.strftime('%H:%M:%S.%f') + time.strftime(' - %d.%m.%Y')
        mes.append({'time': formatted_time, 'content': message, 'user': from_user})  # append message
        json.dump(mes, open(Const.ChatFile, 'w'), indent=4)  # write message

    @staticmethod
    def get() -> list:
        """
        get all messages
        """
        try:
            mes = json.load(open(Const.ChatFile, 'r'))  # try to read file
        except FileNotFoundError:
            mes = list()  # if file doesn't exist, create new list
        return mes


class Communication:
    """
    Handler for server side communication between Server and Client
    """

    @staticmethod
    def send(client: socket.socket, message: dict | list, encryption=None, key=None) -> None:
        """
        send message to client
        """
        string_mes = json.dumps(message, ensure_ascii=False)
        print(string_mes)
        if encryption:
            mes = encryption(string_mes, key=key)
            # print(mes)
            with contextlib.suppress(OSError, AttributeError):
                client.send(mes)
                print(f'sent to client: {mes}')
            print('failed to send')
            return

        with contextlib.suppress(OSError, AttributeError):
            print(f'sent to client ({client}): {string_mes.encode("utf-8")}')
            client.send(string_mes.encode('utf-8'))

    @staticmethod
    def receive(server: socket.socket, debugging_method, keys: list | typing.Generator
                ) -> typing.Tuple[socket.socket, str] | typing.Tuple[None, None] | bool:
        """
        receive message from client
        """
        # Accept Connection
        try:
            client, address = server.accept()
            del address
            # debug.debug(f'Connected to {address}')
        except OSError:
            return False
        # try to load the message, else ignore it and restart
        mes = client.recv(2048)
        mes = cryption_tools.try_decrypt(mes, keys)

        if not mes:
            debugging_method('Message Error')
            with contextlib.suppress(AttributeError):
                Communication.send(client, {'Error': 'MessageError', 'info': 'Invalid Message/AuthKey'})
                client.close()
            return None, None
        return client, mes



"""
for weather-stations to commit data to the pool

Author:
Nilusink
"""
from fridrich.server.server_funcs import send_success
from fridrich.server.classes import User
from fridrich.server import Const
from traceback import format_exc
import json


def register(message: dict, user: User, *_args) -> None:
    """
    register a new weather-station
    """
    tmp: list
    try:
        tmp = json.load(open(Const.WeatherDir+"all.json", "r"))

    except (json.JSONDecodeError, FileNotFoundError):
        tmp = []

    for element in tmp:
        if message["station_name"] == element["station_name"]:
            mes = {
                    'Error': 'RegistryError',
                    "info": "weather-station is already registered"
                }
            user.send(mes, message_type="Error")
            return

    tmp.append({
        "station_name": message["station_name"],
        "location": message["location"]
    })

    with open(Const.WeatherDir+"all.json", "w") as out_file:
        json.dump(tmp, out_file, indent=4)

    with open(Const.WeatherDir+message["station_name"]+".json", "w") as out_file:
        out_file.write("{}")

    send_success(user)


def commit_data(message: dict, user: User, *_args) -> None:
    """
    commit data for already registered stations
    """
    now_data: dict
    station_data: dict
    try:
        if not check_if_registered(message, user, *_args):
            mes = {
                    'Error': 'RegistryError',
                    "info": "weather-station is not registered yet"
                }
            user.send(mes, message_type="Error")
            return

        try:
            now_data = json.load(open(Const.WeatherDir+"now.json", "r"))

        except (json.JSONDecodeError, FileNotFoundError):
            now_data = {}

        station_name = message["station_name"]
        commit_time = message["time"]
        message.pop("type")
        message.pop("station_name")

        now_data[station_name] = message.copy()  # copy so we don't get strange dependencies

        with open(Const.WeatherDir+"now.json", "w") as out_file:
            json.dump(now_data, out_file, indent=4)

        try:
            station_data = json.load(open(Const.WeatherDir+station_name+".json", "r"))

        except (json.JSONEncoder, FileNotFoundError):
            station_data = {}

        message.pop("time")
        station_data[commit_time] = message.copy()

        with open(Const.WeatherDir + station_name+".json", "w") as out_file:
            json.dump(station_data, out_file, indent=4)

        send_success(user)

    except KeyError:
        print(format_exc())
        user.send({
            "Error": "KeyError",
            "info": "Not all keys are given",
            "full": format_exc()
        })


def check_if_registered(message: dict, _user: User, *_args) -> bool:
    """
    check if a weather-station is already registered
    """
    try:
        return message["station_name"] in [station["station_name"] for station in json.load(open(Const.WeatherDir+"all.json", "r"))]

    except FileNotFoundError:
        return False


def get_now(_message: dict, user: User, *_args) -> None:
    """
    send a dict of all weather-stations with their current measurement
    """
    now_data: dict
    try:
        now_data = json.load(open(Const.WeatherDir+"now.json", "r"))

    except (json.JSONDecodeError, FileNotFoundError):
        now_data = {}

    user.send(now_data)


def get_log(message: dict, user: User, *_args) -> None:
    """
    send the log of a specific weather station
    """
    try:
        station_log = json.load(open(Const.WeatherDir+message["station_name"]+".json", "r"))

    except FileNotFoundError:
        user.send({
            "Error": "KeyError",
            "info": f"Weather station with name ({message['station_name']}) not registered!"
        })
        return

    user.send(station_log)


def get_stations(_message: dict, user: User, *_args) -> None:
    """
    send a dict of all currently registered weather station names and locations
    """
    try:
        stations = json.load(open(Const.WeatherDir+"all.json", "r"))

    except FileNotFoundError:
        user.send({})
        return

    user.send(stations)



from fridrich import cryption_tools, decorate_class
from concurrent.futures import ThreadPoolExecutor
from fridrich.server.accounts import USER_CONFIG
from fridrich.classes import Daytime
from fridrich.server import DEBUGGER
from contextlib import suppress
from struct import pack
import socket
import typing
import time
import json


@decorate_class(DEBUGGER.catch_traceback())
class User:
    def __init__(self, name: str, sec: str, key: str, user_id: int, cl: socket.socket, ip: str,
                 function_manager: typing.Callable, debugger) -> None:
        """
        :param name: Name of the client
        :param sec: security clearance
        :param key: encryption key of client
        :param cl: the users' socket instance
        :param ip: the users host ip
        :param function_manager: the manager class for executing functions
        :param debugger: an instance of server_funcs.Debug
        """
        self.__name = name
        self.__sec = sec
        self.__key = key
        self.__id = user_id

        self.__client = cl
        self.__ip = ip
        self.manager = function_manager
        self.debugger = debugger

        self.__disconnect = False

        self.loop = True

        # message pool
        self.__message_pool_names: typing.Tuple[str, str] = ("", "")
        self.__message_pool: typing.Dict[str, typing.Any] = {}
        self.__message_pool_index: int = 0
        self.__message_pool_max: int = 0
        self.__message_pool_time: str = ""

        # for auto-disconnect
        self.__last_connection = Daytime.now()
        # if there wasn't any interaction with a client for 2 minutes, kick it (ping also counts as activity)
        self.__timeout = Daytime(minute=3) 
        self.__thread_pool = ThreadPoolExecutor(max_workers=2)

        # only if the user's security clearance requires auto-logout, start the thread
        if USER_CONFIG[self.sec]["auto_logout"]:
            self.__auto_logout_thread = self.__thread_pool.submit(self.__check_disconnect)

        # start to receive
        self.__thread_pool.submit(self.receive)

    @property
    def name(self) -> str:
        """
        :return: the username of the user
        """
        return self.__name

    @property
    def sec(self) -> str:
        """
        :return: the security clearance of the user
        """
        return self.__sec

    @property
    def ip(self) -> str:
        """
        :return: the users host-ip
        """
        return self.__ip

    @property
    def id(self) -> int:
        """
        :return: the users id
        """
        return self.__id

    @property
    def disconnect(self) -> bool:
        return self.__disconnect

    def receive(self) -> None:
        """
        needs to be run in a thread, handles communication with client
        """
        self.__client.settimeout(.2)
        while self.loop:
            try:
                mes = cryption_tools.MesCryp.decrypt(self.__client.recv(2048), self.__key.encode())

                # refresh auto-disconnect
                self.__last_connection = Daytime.now()

                mes = json.loads(mes)

                # create message pool for request
                names = [func_name["f_name"] if "f_name" in func_name else func_name["type"] 
                         for func_name in mes["content"]]
                self.__message_pool_names = tuple(*names)
                self.__message_pool_max = len(mes["content"])
                self.__message_pool_time = mes["time"]
                self.__message_pool_index = 0

                if not mes or mes is None:
                    self.send({'Error': 'MessageError', 'info': 'Invalid Message/AuthKey'},
                              message_type="Error", force=True)
                    continue

                for message in mes["content"]:
                    self.exec_func(message)

            except cryption_tools.NotEncryptedError:
                print("not encrypted")
                self.send({'Error': 'NotEncryptedError'}, message_type="Error", force=True)
                return

            except (TimeoutError, OSError):
                continue

    def send(self, message: iter, message_type: str | None = 'function', force: bool | None = False) -> None:
        """
        save the message(s) for sending
        """
        message = {
            "content": message
        }
        if self.__message_pool_max == sum([0 if element is None else 1 
                                           for element in self.__message_pool]) and not force:
            print(f"Pool error: {self.__message_pool=}, {self.__message_pool_names=}, {message=}")
            raise IndexError("trying to send message but no pool index is out of range")

        message['type'] = message_type

        try:
            self.__message_pool[self.__message_pool_names[self.__message_pool_index]] = message
            self.__message_pool_index += 1

        # if used with "force", appends the message in case of a IndexError (pool not created)
        except IndexError:
            if force:
                self.__message_pool["forced"] = message
            else:
                raise

        if force or self.__message_pool_index == self.__message_pool_max:   # aka all functions are done
            self._send()

    def _send(self) -> None:
        """
        actually send the message
        """
        # refresh auto-disconnect
        self.__last_connection = Daytime.now()

        # process the message
        mes = {
            "content": self.__message_pool,
            "time": self.__message_pool_time
        }
        string_mes = json.dumps(mes)
        mes = cryption_tools.MesCryp.encrypt(string_mes, key=self.__key.encode())
        length = pack('>Q', len(mes))   # get message length

        # send to client
        try:
            self.__client.sendall(length)
            self.__client.sendall(mes)

        except (OSError, ConnectionResetError, ConnectionAbortedError):
            return self.end()

        # reset message pool
        self.__message_pool = {}
        self.__message_pool_index = 0
        self.__message_pool_max = 0
        self.__message_pool_time = ""
        self.__message_pool_names = ()

    def __check_disconnect(self) -> None:
        """
        for auto-disconnect, check if there was no interaction with the user for self.__timeout
        """
        while self.loop:
            if Daytime.now()-self.__last_connection > self.__timeout:
                print(f"disconnecting {self.name}, last contact: {self.__last_connection},"
                      f"now: {Daytime.now()} ({self.__timeout=})")
                return self.end("timeout")

            time.sleep(.2)

    def exec_func(self, message: dict):
        """
        execute functions for the client
        """
        try:
            self.manager(message, self)

        except Exception as e:
            self.debugger.write_traceback(e, from_user=self.name)
            return

    def end(self, reason: str = ...) -> None:
        print(f"Disconnecting: {self} {'('+reason+')' if reason is not ... else ''}, shutting down threads")
        self.__disconnect = True
        self.loop = False
        self.__client.close()
        self.__thread_pool.shutdown(wait=False)
        print(f"{self} is fully shut down and disconnected")

    def __getitem__(self, item) -> str:
        return dict(self)[item]

    def __iter__(self) -> typing.Iterator:
        for key, item in (('key', self.__key), ('name', self.__name), ('sec', self.__sec)):
            yield key, item

    def __str__(self) -> str:
        return f"<class User (name={self.__name}, sec={self.__sec})>"

    def __repr__(self) -> str:
        return self.__str__()

    def __contains__(self, item) -> bool:
        return item == self.name or item == self.id


@decorate_class(DEBUGGER.catch_traceback(raise_error=False),
                dont_decorate=["_garbage_collector", "__contains__", "get_user"])
class UserList:
    def __init__(self, users: typing.List[User] | None = ...) -> None:
        """
        initialize a list for all users and start garbage-collector

        special: ´´get_user´´ function (gets a user by its name or encryption key)
        """
        self._users = users if users is not ... else list()

        self.executor = ThreadPoolExecutor()
        self.collector = self.executor.submit(self._garbage_collector, .5)

        self.loop = True

    @property
    def names(self) -> typing.Generator:
        """
        return the names of all users
        """
        for element in self._users:
            yield element.name

    def sendall(self, message: iter) -> None:
        for user in self._users:
            user.send({
                "time": time.time(),
                "content": message
            }, force=True)

    def append(self, obj: User) -> None:
        """
        append object to the end of the list and start receive thread
        """
        if obj.name in self and not USER_CONFIG[obj.sec]["multi_login_allowed"]:
            print(f"multi login disallowed, logging out other users")
            with suppress(KeyError):
                self.remove_by(name=obj.name)

        self._users.append(obj)

    def get_user(self, name: str | None = ..., user_id: str | None = ...) -> User:
        """
        get a user by its name or encryption key
        """
        for element in self._users:
            if name is not ...:
                if name in element:
                    return element

            if user_id is not ...:
                if user_id in element:
                    return element

        else:
            raise KeyError(f'No User with name {name} or id {user_id} found!')

    def remove(self, user: User) -> None:
        """
        remove a user by its class
        """
        user.end(reason="deleted from list")
        self._users.remove(user)

    def remove_by(self, *args, **kw) -> None:
        """
        remove a user by its username or encryption key

        arguments are the same as for UserList.get_user
        """
        self.remove(self.get_user(*args, **kw))

    def reset(self) -> None:
        """
        reset all users (clear self._users)
        """
        for user in self._users:
            user.send({'warning': 'server_logout'}, message_type="disconnect", force=True)
            user.end()
        self._users = list()

    def _garbage_collector(self, time_between_loops: float | None = .5) -> None:
        """
        check if any of the clients is disconnected and if yes, remove it
        """
        while self.loop:
            for element in self._users:
                if element.disconnect:
                    print(f"removing {element.name}")
                    element.end()
                    self._users.remove(element)
            time.sleep(time_between_loops)

    def end(self) -> None:
        """
        shutdown all threads
        """
        self.loop = False
        for user in self._users:
            user.end()
        self.executor.shutdown(wait=False)

    def __iter__(self) -> typing.Iterator:
        for element in self._users:
            yield element

    def __contains__(self, other: str) -> bool:
        with suppress(KeyError):
            self.get_user(name=other, user_id=other)
            return True
        return False

    def __str__(self) -> str:
        return f"<class UserList (users={list(self.names)}, running={self.__bool__()})>"

    def __repr__(self) -> str:
        return self.__str__()

    def __bool__(self) -> bool:
        """
        return True if AuthKey
        """
        return self.loop



from fridrich.server.classes import User
from fridrich.server import Const
from time import strftime


def ping(message: dict, user: User, *_args) -> None:
    """
    immediately send back a message (to measure latency)
    """
    user.send(message)


def get_time(_message: dict, user: User, *_args) -> None:
    """
    get the current server time
    """
    user.send({
            "now": strftime("%H:%M:%S"),
            "voting": Const.switchTime
        })


def get_sec_clearance(_message: dict, user: User, *_args):
    """
    send the user its security clearance
    """
    user.send(user.sec)



"""
The fridrich module is a set of scripts used
to drive a Server, which is used mainly for
the Votings (There is much more to this,
but it's not really easy to explain),
and a Client that connects to the Server

Author:
Nilusink
"""


###########################################################################
#                  classes and functions used everywhere                  #
###########################################################################
def decorate_class(decorator, dont_decorate: list = ()):
    """
    decorate all methods of a class with one decorator
    """

    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in dont_decorate:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


class ConsoleColors:
    """
    for better readability (colors) in the console
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



"""
used for any type on en/decryption
for fridrich (ex. End-to-End encryption,
password hashing, private config files
for the Client, ...)
(Server & Client)

Author: Nilusink
"""
import contextlib
import random
import math
import json
import os

# cryptography imports
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives import hashes
import base64


class DecryptionError(Exception):
    pass


class NotEncryptedError(Exception):
    pass


class Extra:
    """
    extra functions
    """

    @staticmethod
    def median(string_: str, medians: int) -> str:
        """
        split in medians number of parts and then reverse
        """
        parts = list()
        out = list()
        for i in range(1, medians + 1):
            if not i == medians:
                parts.append([int((len(string_) - 1) / medians * (i - 1)), int((len(string_) - 1) / medians * i)])
            else:
                parts.append([int((len(string_) - 1) / medians * (i - 1)), len(string_)])
        for part in parts:
            out.append(string_[::-1][part[0]:part[1]])
        return ''.join(out[::-1])


class Low:
    @staticmethod
    def encrypt(string_: str) -> str:
        """
        encrypt a string_
        """
        out = str()
        for charter in string_:
            part = str(math.sqrt(ord(charter) - 20))
            out += str(base64.b85encode(part.encode('utf-16'))).lstrip("b'").rstrip("='") + ' '
        return out

    @staticmethod
    def decrypt(string_: str) -> str:
        """
        decrypt a string_
        """
        try:
            out = str()
            parts = string_.split(' ')
            for part in parts:
                s = (part + '=').encode()
                if not s == b'=':
                    part = float(base64.b85decode(part).decode('utf-16'))
                    out += chr(int(round(part ** 2 + 20, 0)))
            return out
        except ValueError:
            raise DecryptionError('Not a valid encrypted string_!')


class High:
    @staticmethod
    def encrypt(string_: str) -> str:
        """
        encrypt a string_
        """
        temp1, temp2 = str(), str()
        for charter in string_:
            temp1 += Low.encrypt((Extra.median(charter, 3) + ' ')) + ' '
        for charter in Extra.median(temp1, 13):
            temp2 += str(ord(charter)) + '|1|'
        temp2 = Low.encrypt(temp2)
        out = Extra.median(Extra.median(temp2, 152), 72)
        return Extra.median(str(base64.b85encode(out.encode('utf-32'))).lstrip("b'").rstrip("='")[::-1], 327)

    @staticmethod
    def decrypt(string_: str) -> str:
        """
        decrypt a string_
        """
        temp1, temp2 = str(), str()
        string_ = Extra.median(string_, 327)[::-1]
        string_ = base64.b85decode(string_).decode('utf-32')
        string_ = Extra.median(Extra.median(string_, 72), 152)
        string_ = Low.decrypt(string_)
        parts = string_.split('|1|')
        for part in parts:
            with contextlib.suppress(ValueError):
                temp1 += chr(int(part))
        temp1 = Extra.median(temp1, 13)
        parts = temp1.split(' ')
        for part in parts:
            temp2 += Extra.median(Low.decrypt(part), 3)
        return temp2.replace('   ', '|tempspace|').replace(' ', '').replace('|tempspace|', ' ')


try:
    with open(os.getcwd()+'/data/KeyFile.enc', 'r') as inp:
        defKey = Low.decrypt(inp.read()).encode()

except FileNotFoundError:
    raise FileNotFoundError("Cannot find file data/KeyFile.enc")


class MesCryp:
    """
    encryption/decryption for messages
    """

    @staticmethod
    def encrypt(string_: str, key=None) -> bytes:
        """
        encrypt a string_

        if a key is given, use it
        """
        if not key:
            key = defKey
        f = Fernet(key)
        encrypted = f.encrypt(string_.encode('utf-8'))
        return encrypted  # returns bytes

    @staticmethod
    def decrypt(byte: bytes, key: bytes | None = defKey) -> str:
        """
        decrypt a bytes element
        """
        f = Fernet(key)
        decrypted = str(f.decrypt(byte)).lstrip("b'").rstrip("'")
        return decrypted  # returns string_


def try_decrypt(message: bytes, client_keys: dict | list, errors=True) -> str | None:
    """
    try to decrypt a string_ with multiple methods
    """
    with contextlib.suppress(json.JSONDecodeError):
        mes = json.loads(message)
        if errors:
            raise NotEncryptedError('Message not encrypted')
        print(mes)
        return mes

    enc_mes = None
    for key in client_keys:
        with contextlib.suppress(InvalidToken, ValueError):
            enc_mes = MesCryp.decrypt(message, key.encode() if type(key) == str else b'')
            break

    if not enc_mes:
        with contextlib.suppress(InvalidToken):
            enc_mes = MesCryp.decrypt(message, defKey)

    if not enc_mes:
        print(enc_mes)
        print(message)
        return None

    try:
        json_mes = json.loads(enc_mes)

    except json.JSONDecodeError:
        try:
            json_mes = json.loads(message)

        except json.JSONDecodeError:
            return None
    return json_mes


def key_func(length=10) -> str:
    """
    generate random key
    """
    string = 'abcdefghijklmnopqrstuvwxyz'  # string for creating auth Keys
    string += string.upper() + '1234567890ß´^°!"§$%&/()=?`+*#.:,;µ@€<>|'

    password_provided = ''.join(random.sample(string, length))  # This is input in the form of a string_
    password = password_provided.encode()  # Convert to type bytes
    salt = os.urandom(16)
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )
    key = base64.urlsafe_b64encode(kdf.derive(password))  # Can only use kdf once
    return str(key).lstrip("b'").rstrip("'")
